import requests
import asyncio
import aiohttp
import json

class OsrmEngine:
    def __init__(
            self,
            osrm_url:str,
            verified_url:bool=True,
            profile:str="driving",
            version:str="v1"
        ):
        self._osrm_url = osrm_url
        self._verified_url = verified_url
        self._profile = profile
        self._version = version

    # Getters and setters
    @property
    def osrm_url(self): return self._osrm_url
    @property
    def verified_url(self): return self._verified_url
    @property
    def profile(self): return self._profile
    @property
    def version(self): return self._version

    def get_nearest(self, location:dict, number:int=1)->list:
        url = f"{self.osrm_url.removesuffix('/')}/nearest/{self.version}/{self.profile}/{location['lng']},{location['lat']}"
        url += f"?number={number}"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        data = response.json()
        if data["code"] != "Ok":
            raise Exception(f"Error {data['code']}: {data['message']}")
        return data["waypoints"]

    def get_route(self, start, end,
        alternatives:bool=False,
        steps:bool=False,
        annotations:bool=False,
        geometries:str="polyline",
        overview:str="simplified",
        continue_straight:str="default"
    )->list:
        url = f"{self.osrm_url.removesuffix('/')}/route/{self.version}/{self.profile}/{start['lng']},{start['lat']};{end['lng']},{end['lat']}"
        # Properties
        url += f"?alternatives={str(alternatives).lower()}"
        url += f"&steps={str(steps).lower()}"
        url += f"&annotations={str(annotations).lower()}"
        url += f"&geometries={geometries}"
        url += f"&overview={overview}"
        url += f"&continue_straight={continue_straight}"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        data = response.json()
        if data["code"] != "Ok":
            raise Exception(f"Error {data['code']}: {data['message']}")
        return data["routes"]
    
    def get_route_async(
        self,
        start:list,
        end:list,
        alternatives:bool=False,
        steps:bool=False,
        annotations:bool=False,
        geometries:str="polyline",
        overview:str="simplified",
        continue_straight:str="default"
    )->list:
        # Check start and end have the same length
        if len(start) != len(end):
            raise Exception("Start and end must have the same length")
        # Prepare urls
        urls = []
        for i in range(len(start)):
            url = f"{self.osrm_url.removesuffix('/')}/route/{self.version}/{self.profile}/{start[i]['lng']},{start[i]['lat']};{end[i]['lng']},{end[i]['lat']}"
            # Properties
            url += f"?alternatives={str(alternatives).lower()}"
            url += f"&steps={str(steps).lower()}"
            url += f"&annotations={str(annotations).lower()}"
            url += f"&geometries={geometries}"
            url += f"&overview={overview}"
            url += f"&continue_straight={continue_straight}"
            urls.append(url)
        # Async
        async def get(url, session):
            try:
                async with session.get(url=url) as response:
                    resp = await response.read()
                    return resp
                    # print("Successfully got url {} with resp of length {}.".format(url, len(resp)))
            except Exception as e:
                # print("Unable to get url {} due to {}.".format(url, e.__class__))
                pass
        async def main(urls):
            async with aiohttp.ClientSession() as session:
                ret = await asyncio.gather(*(get(url, session) for url in urls))
                return ret
            # print("Finalized all. Return is a list of len {} outputs.".format(len(ret)))
            # return ret
        # Run
        data = asyncio.run(main(urls))
        return [d['routes'] for d in [json.loads(d) for d in data]]

        
    
    def get_table(self, locations:list, sources:list=None, destinations:list=None)->list:
        url = f"{self.osrm_url.removesuffix('/')}/table/{self.version}/{self.profile}/"
        url += f";".join([f"{location['lng']},{location['lat']}" for location in locations])
        if sources or destinations:
            url += "?"
        if sources:
            url += f"sources={';'.join([str(source) for source in sources])}"
        if destinations:
            if sources:
                url += "&"
            url += f"destinations={';'.join([str(destination) for destination in destinations])}"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        data = response.json()
        if data["code"] != "Ok":
            raise Exception(f"Error {data['code']}: {data['message']}")
        del data["code"]
        return data
    
    def get_match(
        self,
        locations:list,
        steps:bool=False,
        geometries:str="polyline",
        annotations:bool=False,
        overview:str="simplified",
        timestamps:list=None,
        radiuses:list=None
    )->list:
        url = f"{self.osrm_url.removesuffix('/')}/match/{self.version}/{self.profile}/"
        url += f";".join([f"{location['lng']},{location['lat']}" for location in locations])
        # Properties
        url += f"?steps={str(steps).lower()}"
        url += f"&geometries={geometries}"
        url += f"&annotations={str(annotations).lower()}"
        url += f"&overview={overview}"
        if timestamps:
            url += f"&timestamps={';'.join([str(timestamp) for timestamp in timestamps])}"
        if radiuses:
            url += f"&radiuses={';'.join([str(radius) for radius in radiuses])}"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        data = response.json()
        if data["code"] != "Ok":
            raise Exception(f"Error {data['code']}: {data['message']}")
        del data["code"]
        return data
    
    def get_trip(
        self,
        locations:list,
        steps:bool=False,
        annotations:bool=False,
        geometries:str="polyline",
        overview:str="simplified",
    ):
        url = f"{self.osrm_url.removesuffix('/')}/trip/{self.version}/{self.profile}/"
        url += f";".join([f"{location['lng']},{location['lat']}" for location in locations])
        # Properties
        url += f"?steps={str(steps).lower()}"
        url += f"&annotations={str(annotations).lower()}"
        url += f"&geometries={geometries}"
        url += f"&overview={overview}"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        data = response.json()
        if data["code"] != "Ok":
            raise Exception(f"Error {data['code']}: {data['message']}")
        del data["code"]
        return data
    
    def get_tile(
        self,
        x:int,
        y:int,
        zoom:int
    )->bytes:
        url = f"{self.osrm_url.removesuffix('/')}/tile/{self.version}/{self.profile}/tile({x},{y},{zoom}).mvt"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        return response.content
    
    



import pandas as pd
from sqlalchemy import text

from galigeopy.model.zone import Zone

class ZoneType:
    def __init__(
        self,
        zone_type_id:int,
        name:str,
        description:str,
        org: 'Org' # type: ignore
    ):
        # Infos
        self._zone_type_id = zone_type_id
        self._name = name
        self._description = description
        # Engine
        self._org = org

    def __str__(self):
        return self.name + " (" + str(self.zone_type_id) + ")"
    
    # Getters and setters
    @property
    def zone_type_id(self): return self._zone_type_id
    @property
    def name(self): return self._name
    @property
    def description(self): return self._description
    @property
    def org(self): return self._org
    
    # Public Methods
    def number_of_zones(self):
        query = text(f"SELECT COUNT(*) FROM ggo_zone WHERE zone_type_id = {self.zone_type_id}")
        with self._org.engine.connect() as conn:
            result = conn.execute(query)
            return result.scalar()
        
    def getZonesList(self):
        query = text(f"SELECT * FROM ggo_zone WHERE zone_type_id = {self.zone_type_id}")
        return pd.read_sql(query, self._org.engine)
    
    def getZoneById(self, zone_id:int):
        query = f"SELECT * FROM ggo_zone WHERE zone_id = {zone_id} AND zone_type_id = {self.zone_type_id}"
        df = pd.read_sql(query, self._org.engine)
        if len(df) > 0:
            data = df.iloc[0].to_dict()
            data.update({"org": self._org})
            return Zone(**data) 
        else:
            raise Warning(f"Zone {zone_id} not found in ZoneType {self.name}")
        
    def getAllZones(self):
        query = text(f"SELECT * FROM ggo_zone WHERE zone_type_id = {self.zone_type_id}")
        df = pd.read_sql(query, self._org.engine)
        zones = []
        for i in range(len(df)):
            data = df.iloc[i].to_dict()
            data.update({"org": self._org})
            zones.append(Zone(**data))
        return zones
        
    def add_to_model(self)-> int:
        # Add to database
        query = f"""
        INSERT INTO ggo_zone_type (
            name
        ) VALUES (
            '{self.name.replace("'", "''")}'
        ) RETURNING zone_type_id;
        """
        zone_type_id = self._org.query(query)[0][0]
        return zone_type_id




import unittest
import json

from galigeopy.org.org import Org
from galigeopy.model.network import Network
from galigeopy.model.poi import Poi

class TestNetwork(unittest.TestCase):
    
    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self.conf = json.load(open("test-config.json"))
        self.org = Org(**self.conf["org"])
        self.network = self.org.getNetworkById(self.conf["network_id"])

    def test_poi(self):
        poi = self.network.getPoiByCode(self.conf["poi_code"])
        self.assertIsNotNone(poi)
        self.assertIsInstance(poi, Poi)
        self.assertEqual(poi.id, self.conf["poi_code"])
        self.assertIsNotNone(poi.x)


import unittest
import json

from galigeopy.org.org import Org
from galigeopy.model.zone import Zone
from galigeopy.model.poi import Poi

class TestZone(unittest.TestCase):

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self.conf = json.load(open("test-config.json"))
        self.org = Org(**self.conf["org"])
        self.zone_type = self.org.getZoneTypeById(self.conf["zone_type_id"])

    def test_zone(self):
        zone = self.zone_type.getZoneById(self.conf["zone_id"])
        self.assertIsNotNone(zone)
        self.assertIsInstance(zone, Zone)
        self.assertEqual(zone.zone_id, self.conf["zone_id"])
        del zone

    def test_get_poi(self):
        zone = self.zone_type.getZoneById(self.conf["zone_id"])
        poi = zone.getPoi()
        self.assertIsNotNone(poi)
        self.assertIsInstance(poi, Poi)
        self.assertEqual(poi.poi_id, zone.poi_id)
        del zone

    def test_get_parent_zone(self):
        # TODO: Implement test
        pass

    def test_get_child_zones(self):
        # TODO: Implement test
        pass

    def test_get_zone_geounit_list(self):
        zone = self.zone_type.getZoneById(self.conf["zone_id"])
        df = zone.getZoneGeounitsList()
        self.assertIsNotNone(df)
        self.assertGreater(len(df), 0)
        self.assertEqual(df.iloc[0]["zone_id"], self.conf["zone_id"])
        del zone

    def test_get_zone_geounit_by_id(self):
        zone = self.zone_type.getZoneById(self.conf["zone_id"])
        zone_geounit = zone.getZoneGeounitById(self.conf["zone_geounit_id"])



import unittest
import json

from galigeopy.org.org import Org
from galigeopy.model.zone_type import ZoneType
from galigeopy.model.zone import Zone
from galigeopy.model.poi import Poi
from galigeopy.model.zone_geounit import ZoneGeounit

class TestZoneType(unittest.TestCase):

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self.conf = json.load(open("test-config.json"))
        self.org = Org(**self.conf["org"])
        self.zone_type = self.org.getZoneTypeById(self.conf["zone_type_id"])
        self.zone = self.zone_type.getZoneById(self.conf["zone_id"])
    
    def test_zone_geounit(self):
        zone_geounit = self.zone.getZoneGeounitById(self.conf["zone_geounit_id"])
        self.assertIsNotNone(zone_geounit)
        self.assertIsInstance(zone_geounit, ZoneGeounit)
        self.assertEqual(zone_geounit.zone_geounit_id, self.conf["zone_geounit_id"])
        del zone_geounit
    
    def test_get_zone_geounits_list(self):
        df = self.zone.getZoneGeounitsList()
        self.assertIsNotNone(df)
        self.assertGreater(len(df), 0)
        self.assertEqual(df.iloc[0]["zone_id"], self.conf["zone_id"])

    # def test_get_zone (self):
    #     zone = self.zone_geounit.getZone()
    #     self.assertIsNotNone(zone)
    #     self.assertIsInstance(zone, Zone)
    #     self.assertEqual(zone.zone_id, self.conf["zone_id"])
    #     del zone


import pandas as pd
import geopandas as gpd
from sqlalchemy import text

from .geolevel import Geolevel
from .network import Network

class DistancierSession:
    # Constructor
    def __init__(
            self,
            session_id:int,
            name:str,
            engine:str,
            network_id_start:int,
            network_id_end:int,
            geolevel_id_start:int,
            geolevel_id_end:int,
            direction:str,
            max_distance:int,
            max_time:int,
            max_calc_out:int,
            no_route:bool,
            org:'Org' # type: ignore
    ):
        self._session_id = session_id
        self._name = name
        self._engine = engine
        try:
            self._network_id_start = int(network_id_start)
        except: 
            self._network_id_start = None
        try:
            self._network_id_end = int(network_id_end)
        except:
            self._network_id_end = None
        try:
            self._geolevel_id_start = int(geolevel_id_start)
        except:
            self._geolevel_id_start = None
        try:
            self._geolevel_id_end = int(geolevel_id_end)
        except:
            self._geolevel_id_end = None
        self._direction = direction
        try:
            self._max_distance = int(max_distance)
        except:
            self._max_distance = None
        try:
            self._max_time = int(max_time)
        except:
            self._max_time = None
        try:
            self._max_calc_out = int(max_calc_out)
        except:
            self._max_calc_out = None
        self._no_route = no_route
        self._org = org

    # Getters and setters
    @property
    def session_id(self): return self._session_id
    @property
    def name(self): return self._name
    @property
    def engine(self): return self._engine
    @property
    def network_id_start(self): return self._network_id_start
    @property
    def network_id_end(self): return self._network_id_end
    @property
    def geolevel_id_start(self): return self._geolevel_id_start
    @property
    def geolevel_id_end(self): return self._geolevel_id_end
    @property
    def direction(self): return self._direction
    @property
    def max_distance(self): return self._max_distance
    @property
    def max_time(self): return self._max_time
    @property
    def max_calc_out(self): return self._max_calc_out
    @property
    def no_route(self): return self._no_route
    @property
    def org(self): return self._org

    # Magic Method
    def __str__(self) -> str:
        return f"DistancierSession({self._session_id} - {self._name})"
    
    # Public Methods
    def getStartingType(self)->str:
        if self._network_id_start:
            return 'network'
        else:
            return 'geolevel'
    def getEndingType(self)->str:
        if self._network_id_end:
            return 'network'
        else:
            return 'geolevel'
    def getStartingObject(self)->Network|Geolevel:
        if self._network_id_start:
            return self._org.getNetwork(self._network_id_start)
        else:
            return self._org.getGeolevel(self._geolevel_id_start)
        
    def getEndingObject(self)->Network|Geolevel:
        if self._network_id_end:
            return self._org.getNetwork(self._network_id_end)
        else:
            return self._org.getGeolevel(self._geolevel_id_end)
        
    def getDistancier(self)->pd.DataFrame:
        # Query
        q = f"SELECT * FROM ggo_distancier WHERE session_id = {self._session_id}"
        # Df
        return self._org.query_df(q)
    
    def to_json(self):
        return {
            "session_id": self._session_id,
            "name": self.name,
            "engine": self._engine,
            "network_id_start": self._network_id_start,
            "network_id_end": self._network_id_end,
            "geolevel_id_start": self._geolevel_id_start,
            "geolevel_id_end": self._geolevel_id_end,
            "direction": self._direction,
            "max_distance": self._max_distance,
            "max_time": self._max_time,
            "max_calc_out": self._max_calc_out,
            "no_route": self._no_route
        }
    
    def getPoisIdList(self)->list:
        pois_ids_list = []
        if self.getStartingType() == 'network':
            q = f"SELECT DISTINCT poi_id_start FROM ggo_distancier WHERE session_id = {self._session_id} AND poi_id_start IS NOT NULL"
            df = self._org.query_df(q)
            pois_ids_list += df['poi_id_start'].tolist()
        if self.getEndingType() == 'network':
            q = f"SELECT DISTINCT poi_id_end FROM ggo_distancier WHERE session_id = {self._session_id} AND poi_id_end IS NOT NULL"
            df = self._org.query_df(q)
            pois_ids_list += df['poi_id_end'].tolist()
        return pois_ids_list
    
    def getGeounitCodeList(self)->list:
        geounit_code_list = []
        if self.getEndingType() == 'geolevel':
            q = f"SELECT DISTINCT geounit_code_start FROM ggo_distancier WHERE session_id = {self._session_id} AND geounit_code_start IS NOT NULL"
            df = self._org.query_df(q)
            geounit_code_list += df['geounit_code_start'].tolist()
        if self.getEndingType() == 'geolevel':
            q = f"SELECT DISTINCT geounit_code_end FROM ggo_distancier WHERE session_id = {self._session_id} AND geounit_code_end IS NOT NULL"
            df = self._org.query_df(q)
            geounit_code_list += df['geounit_code_end'].tolist()
        return geounit_code_list


import unittest
import json
import pandas as pd
import geopandas as gpd

from galigeopy.org.org import Org
from galigeopy.model.geolevel import Geolevel

class TestGeolevel(unittest.TestCase):

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self.conf = json.load(open("test-config.json"))
        self.org = Org(**self.conf["org"])

    def test_geolevel(self):
        geolevel = self.org.getGeolevelById(self.conf["geolevel_id"])
        self.assertIsNotNone(geolevel)
        self.assertIsInstance(geolevel, Geolevel)
        self.assertEqual(geolevel.geolevel_id, self.conf["geolevel_id"])
        del geolevel

    def test_get_geo_dataset(self):
        geolevel = self.org.getGeolevelById(self.conf["geolevel_id"])
        dataset = geolevel.getGeoDataset()
        self.assertIsNotNone(dataset)
        self.assertIsInstance(dataset, pd.DataFrame)
        self.assertGreater(len(dataset), 0)
        del geolevel, dataset

    def test_get_socio_demo_dataset(self):
        geolevel = self.org.getGeolevelById(self.conf["geolevel_id"])
        dataset = geolevel.getSocioDemoDataset()
        self.assertIsNotNone(dataset)
        self.assertIsInstance(dataset, pd.DataFrame)
        self.assertGreater(len(dataset), 0)
        del geolevel, dataset

    def test_get_geo_socio_demo_dataset(self):
        geolevel = self.org.getGeolevelById(self.conf["geolevel_id"])
        dataset = geolevel.getGeoSocioDemoDataset()
        self.assertIsNotNone(dataset)
        self.assertIsInstance(dataset, gpd.GeoDataFrame)
        self.assertGreater(len(dataset), 0)
        del geolevel, dataset




import pandas as pd
import geopandas as gpd
from shapely import wkt
from sqlalchemy import text

from galigeopy.model.poi import Poi

class Network:
    # Constructor
    def __init__(
            self,
            network_id:int,
            org_id:int,
            name:str,
            brand:str,
            network_type_id:str,
            is_default:bool,
            created_at:str,
            created_by:str,
            last_updated_at:str,
            last_updated_by:str,
            geolevel_id:int,
            description:str,
            org:'Org' # type: ignore
    ):
        # Infos
        self._network_id = network_id
        self._org_id = org_id
        self._name = name
        self._brand = brand
        self._network_type_id = network_type_id
        self._is_default = is_default
        self._created_at = created_at
        self._created_by = created_by
        self._last_updated_at = last_updated_at
        self._last_updated_by = last_updated_by
        self._geolevel_id = geolevel_id
        self._description = description
        self._org = org

    # Getters and setters
    @property
    def network_id(self): return self._network_id
    @property
    def org_id(self): return self._org_id
    @property
    def name(self): return self._name
    @property
    def brand(self): return self._brand
    @property
    def network_type_id(self): return self._network_type_id
    @property
    def is_default(self): return self._is_default
    @property
    def created_at(self): return self._created_at
    @property
    def created_by(self): return self._created_by
    @property
    def last_updated_at(self): return self._last_updated_at
    @property
    def last_updated_by(self): return self._last_updated_by
    @property
    def geolevel_id(self): return self._geolevel_id
    @property
    def description(self): return self._description
    @property
    def org(self): return self._org

    # Magic Methods
    def __str__(self):
        return f"Network({self._network_id}, {self._name}, {self._brand})"
    
    # Public Methods
    def number_of_pois(self)->int:
        query = text(f"SELECT COUNT(*) FROM ggo_poi WHERE network_id = {self._network_id}")
        with self._org.engine.connect() as conn:
            result = conn.execute(query)
            return result.scalar()

    def getPoisList(self)->gpd.GeoDataFrame:
        # Query
        query = f"SELECT * FROM ggo_poi WHERE network_id = {self._network_id}"
        # Get data from query
        gdf = gpd.read_postgis(query, self._org.engine, geom_col="geom")
        # return df
        return gdf
    
    def getPoiByCode(self, code:str)->Poi:
        # Query
        query = f"SELECT * FROM ggo_poi WHERE id = '{code}'"
        gdf = gpd.read_postgis(query, self._org.engine, geom_col="geom")
        # Data
        if len(gdf) > 0:
            data = gdf.iloc[0].to_dict()
            data.update({"org": self.org})
            return Poi(**data)
        else:
            raise Warning(f"Poi with code {code} not found in Network {self._name}")
        
    def getAllPois(self)->list:
        # Query
        query = f"SELECT * FROM ggo_poi WHERE network_id = {self._network_id}"
        gdf = gpd.read_postgis(query, self._org.engine, geom_col="geom")
        # Data
        pois = []
        for i in range(len(gdf)):
            data = gdf.iloc[i].to_dict()
            data.update({"org": self.org})
            pois.append(Poi(**data))
        return pois
    
    def getNetworkProperties(self)->pd.DataFrame:
        # Query
        query = f"SELECT properties FROM ggo_poi WHERE network_id = {self._network_id}"
        list_prop = pd.read_sql(query, self._org.engine)["properties"].to_list()
        df_properties = pd.DataFrame(list_prop, dtype='str')
        # Auto detect dtypes
        for col in df_properties.columns:
            df_properties[col] = pd.to_numeric(df_properties[col], errors='ignore')
        df_prop = pd.DataFrame()
        df_prop["columns"] = df_properties.columns
        df_prop["dtypes"] = df_properties.dtypes
        return df_prop
        


from .geolevel_data_type import GeolevelDataType
from .geolevel import Geolevel
from .network import Network
from .poi import Poi
from .zone_geounit import ZoneGeounit
from .zone_type import ZoneType
from .zone import Zone
from .distancier_session import DistancierSession


import unittest
import json

from galigeopy.org.org import Org
from galigeopy.model.zone_type import ZoneType

class TestZoneType(unittest.TestCase):

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self.conf = json.load(open("test-config.json"))
        self.org = Org(**self.conf["org"])

    def test_zone_type(self):
        # Valid ZoneType
        zone_type = self.org.getZoneTypeById(self.conf["zone_type_id"])
        self.assertIsNotNone(zone_type)
        self.assertIsInstance(zone_type, ZoneType)
        self.assertEqual(zone_type.zone_type_id, self.conf["zone_type_id"])
        del zone_type

    def test_number_of_zones(self):
        # Valid ZoneType
        zone_type = self.org.getZoneTypeById(self.conf["zone_type_id"])
        number_of_zones = zone_type.number_of_zones()
        self.assertIsNotNone(number_of_zones)
        self.assertIsInstance(number_of_zones, int)
        self.assertGreater(number_of_zones, 0)
        del zone_type

    def test_get_zones_list(self):
        # Valid ZoneType
        zone_type = self.org.getZoneTypeById(self.conf["zone_type_id"])
        df = zone_type.getZonesList()
        self.assertIsNotNone(df)
        self.assertGreater(len(df), 0)
        self.assertEqual(df.iloc[0]["zone_type_id"], self.conf["zone_type_id"])
        del zone_type